#include "raycasting.h"
#include "scene.h"

Vec3 lighting(Vec3 V, Vec3 N, Vec3 L) { //view,normal,light
    //material
    const Scalar specular_exponent = 80; //(Phong), più è alto il fattore più fitti/densi/piccoli e concentrati i riflessi
    const Vec3 diffuse_color(0.5, 1, 0.5);
    const Vec3 specular_color(1.5, 1.5, 1.5); //"più bianco del bianco"

    //light description
    const Scalar light_intensity = 0.5;

    //lighting "vero e proprio"
    Vec3 H = nlerp(L, V, 0.5);
    Scalar diffuse_factor = std::max(dot(N, L), 0.0); //se negativo dev'essere 0, coefficente di Lambert/diffuse
    Scalar specular_factor = std::pow(std::max(dot(N, H), 0.0), specular_exponent); 

    return (diffuse_factor*light_intensity)*diffuse_color+
           (specular_factor*light_intensity)*specular_color;

}

Picture render(int res, const Scene &scene) {
    const Sphere s = scene.spheres.at(0); //at controlla che ci sia almeno un elemento (perchè 0 è il primo)->range check
    
    const Scalar focal_length = 1; 
    Vec3 light_pos = Vec3(3, 5, 6); //-z -> alle spalle

    Picture image(res, res);
    const Vec3 POV(0, 0, 0);
    const Vec3 background(0, 0, 0); //nero
    
    for (int y = 0; y < image.res_y; y++) 
        for (int x = 0; x < image.res_x; x++) {
            //q è il punto sul piano immagine che corrisponde al punto x,y
            Vec3 q(
                (x - image.res_x/2)/(image.res_x/2.0),
                (y - image.res_y/2)/(image.res_y/2.0),
                focal_length
            );
            Ray r(POV, q - POV);
            Scalar k_min = MISS; 
            int hit_sphere = -1;
            for(int i = 0; i < scene.spheres.size(); i++){
                Scalar k = r.intersect(scene.spheres[i]);
                if(k < k_min) {
                    k_min = k;
                    hit_sphere = i;
                }
            }
            //image.pixel(x, y) = (k==MISS) ? background : colorSphere;

            if(k_min == MISS) {
                image.pixel(x, y) = background;
            }else{
                Vec3 hitpos = r.pos(k_min);
                image.pixel(x, y) = lighting( //diffuse color->in base alla luce, base color->il suo vero colore
                    -r.direction, //- perchè V va verso il punto di vista nel lightning, L va verso la luce e N lontano dalla superficie
                    scene.spheres.at(hit_sphere).normal_of(hitpos),
                    (light_pos - hitpos).normalized() //da dove arriva la luce
                );
            }
        }

    return image;
}